﻿-- clcInfo_ArmsWarrior module.lua -- Deronda
-- don't load if class is wrong
local _, class = UnitClass("player")
if class ~= "WARRIOR" then return end

if not clcInfo then end
local GetTime = GetTime
local debug = clcInfo.debug

local version = 1.2 --bump this if db changes

local modName = "____arms_warrior"
local mod = clcInfo:RegisterClassModule(modName)
local emod = clcInfo.env
local db
local target_health = 100

local db_default = {
	char_var = false,
}

local tdb_default = {
	version = version,
    prio = "tfb15 shout ms2030 exebtrance rend cs exedcreck ms op exe slam",
	ignore_rend_op_less_than_20 = false,
}
-- @defines
--------------------------------------------------------------------------------
local gcdId 				= 85256 -- tv for gcd
-- list of spellId

local rendId				= 772	-- rend
local csId                  = 86346 -- colossus smash
local msId                  = 12294 -- mortal strike
local opId                  = 7384  -- overpower
local exeId                 = 5308  -- execute
local slamId                = 1464  -- slam
local bshoutId              = 6673  -- battle shout
local cshoutId              = 469   -- commanding shout
local tfbId                 = 56638 -- taste for blood
local hsId                  = 78    -- heroic strike
local btId                  = 12964 -- battle trance
local recklessId            = 1719  -- recklessness
local dcId                  = 85730 -- deadly calm
-- str/agil buffs
local howId                 = 57330 -- horn of winter
local rocId                 = 93435 -- roar of courage
local soeId                 = 8075  -- strength of earth totem
-- stam buffs
local qfId                  = 90364 -- qiraji fortitude
local pwfId                 = 21562 -- power word: fortitude
local bpId                  = 6307  -- blood pact

-- devName to pass as argument for melee range checks
local slamName			= GetSpellInfo(slamId)
-- buffs

--Spells for localization
local spells ={
	["Rend"]=rendId,
    ["Colossus Smash"]=csId,
    ["Mortal Strike"]=msId,
    ["Overpower"]=opId,
    ["Execute"]=exeId,
    ["Slam"]=slamId,
    ["Battle Shout"]=bshoutId,
    ["Commanding Shout"]=cshoutId,
    ["Taste for Blood"]=tfbId,
    ["Heroic Strike"]=hsId,
    ["Battle Trance"]=btId,
    ["Recklessness"]=recklessId,
    ["Deadly Calm"]=dcId,
    ["Horn of Winter"]=howId,
    ["Roar of Courage"]=rocId,
    ["Strength of Earth"]=soeId,
    ["Qiraji Fortitude"]=qfId,
    ["Power Word: Fortitude"]=pwfId,
    ["Blood Pact"]=bpId,
}

--localize spells
local name
for key, value in pairs(spells) do
	local name = GetSpellInfo(value)
	if not name then name = "Unknown Spell" end
	spells[key] = { id = value, name = name }
end

-- status vars
local s1, s2
-- current time, old time, remaining gcd, current gcd duration
local s_ctime, s_gcd, s_gcdDur

local shout_id = bshoutId

local function CalculateCurrentTime()
    return s_ctime + s_gcd + s_gcdDur
end

local function CalculateTargetHealthPct()
    return floor(UnitHealth("target") / UnitHealthMax("target") * 100)
end

function mod:OnTemplatesUpdate()
	db = clcInfo:RegisterClassModuleTDB(modName, tdb_default)
	if db then
		if not db.version or db.version < version then
        -- fix stuff
			clcInfo.AdaptConfigAndClean(modName, db, default)
			db.version = version
		end
		
		mod:UpdateQueue()
	end
end


function mod.OnInitialize()
	db = clcInfo:RegisterClassModuleDB(modName, db_default)
end

-- the queue
local q = {}

local function GetCooldown(id)
	local start, duration = GetSpellCooldown(id)
	local cd = start + duration - s_ctime - s_gcd
	if cd < 0 then return 0 end
	return cd
end

local function PickShout()
    --if I cast battle shout, cast it again
    local battle_shout = UnitAura("player", spells["Battle Shout"].name, nil, "HELPFUL|PLAYER")
    if battle_shout then
        return bshoutId
    end
    --if I cast commanding shout, cast it again
    local commanding_shout = UnitAura("player", spells["Commanding Shout"].name, nil, "HELPFUL|PLAYER")
    if commanding_shout then
        return cshoutId
    end
    
    --no shout self buffs... check if anybody else has me buffed.
    --if there is no str+agil buff cast battle shout
    local how = UnitAura("player", spells["Horn of Winter"].name, nil, "HELPFUL")
    local roc = UnitAura("player", spells["Roar of Courage"].name, nil, "HELPFUL")
    local soe = UnitAura("player", spells["Strength of Earth"].name, nil, "HELPFUL")
    local bs = UnitAura("player", spells["Battle Shout"].name, nil, "HELPFUL")    
    if how == nil and roc == nil and soe == nil and bs == nil then
        return bshoutId
    end
    
    --There is a str+agil buff... if there is no stamina buff cast commanding.
    local qf = UnitAura("player", spells["Qiraji Fortitude"].name, nil, "HELPFUL")
    local pwf = UnitAura("player", spells["Power Word: Fortitude"].name, nil, "HELPFUL")
    local bp = UnitAura("player", spells["Blood Pact"].name, nil, "HELPFUL")
    local cshout = UnitAura("player", spells["Commanding Shout"].name, nil, "HELPFUL")
    if qf == nil and pwf == nil and bp == nil and cshout == nil then
        return cshoutId
    end
    
    --sure we could just default to commanding, but we're dps warriors, darn it, so no!!
    return bshoutId --everything is covered so just cast battle
end

-- actions ---------------------------------------------------------------------
local actions = {
	tfb15 = {
		id = function()
            return opId
        end,
		GetCD = function()
            if s1 ~= opId then
                local op_ready, _, _, _, _, _, expirationTime
                     = UnitAura("player", spells["Taste for Blood"].name, nil, "HELPFUL|PLAYER")
                if op_ready ~= nil then
                    local tfb_left = expirationTime - s_ctime - s_gcd
                    local op_cd = GetCooldown(opId)
                    if tfb_left <= 1.5 and op_cd <= 1.5 then
                        return op_cd
                    end
                end
            end
            return 100
        end,
		UpdateStatus = function()
            target_health = CalculateTargetHealthPct()
			s_ctime = CalculateCurrentTime()
		end,
		info = format([[%s: %s <= 1.5]], spells["Overpower"].name, GetSpellInfo(tfbId)),
	},
	shout = {
		id = function()
            return PickShout()
        end,
		GetCD = function()
	        if s1 ~= bshoutId then
                local self_buffed, _, _, _, _, _, expirationTime
                       = UnitAura("player", spells["Battle Shout"].name, nil, "HELPFUL|PLAYER")
                if self_buffed == nil then
                    self_buffed, _, _, _, _, _, expirationTime
                       = UnitAura("player", spells["Commanding Shout"].name, nil, "HELPFUL|PLAYER")
                end
                local self_buffed_time_left = 0
                if self_buffed ~= nil then -- recast our buff if it's about to drop
                    self_buffed_time_left = expirationTime - s_ctime - s_gcd
                end
                if UnitPower("player") < 20 or self_buffed_time_left < 3.0 then
                    return GetCooldown(bshoutId)
                end
            end
            return 100
		end,
		UpdateStatus = function()
            target_health = CalculateTargetHealthPct()
			s_ctime = CalculateCurrentTime()
		end,
		info = format([[%s | %s]], spells["Battle Shout"].name, spells["Commanding Shout"].name),
	},
    ms2030 = {
		id = function()
            return msId
        end,
		GetCD = function()
            if s1 ~= msId and target_health > 20 and UnitPower("player") > 29 then
                return GetCooldown(msId)
            end
            return 100
		end,
		UpdateStatus = function()
            target_health = CalculateTargetHealthPct()
			s_ctime = CalculateCurrentTime()
		end,
		info = format([[%s: 20%% health & > 30 rage]], spells["Mortal Strike"].name),
    },
	exebtrance = {
		id = function()
            return exeId
        end,
		GetCD = function()
			if IsUsableSpell(exeId) then
                local hs_cd = GetCooldown(hsId)
                local exe_cd = GetCooldown(exeId)
                local battle_trance = UnitAura("player", spells["Battle Trance"].name, nil, "HELPFUL|PLAYER")
                if (hs_cd > exe_cd) and battle_trance ~= nil then
                    return exe_cd
                end
			end
			return 100
		end,
		UpdateStatus = function()
            target_health = CalculateTargetHealthPct()
			s_ctime = CalculateCurrentTime()
		end,
		info = format([[%s: %s, %s on cd]], spells["Execute"].name, spells["Battle Trance"].name, spells["Heroic Strike"].name),
	},
    rend = {
		id = function() 
            return rendId
        end,
		GetCD = function()
            if db.ignore_rend_op_less_than_20 == true and target_health < 20 then
                return 100
            end
            local visible = UnitAura("target", spells["Rend"].name, nil, "HARMFUL|PLAYER")
            if visible == nil then
			    if s1 ~= rendId then
				    return GetCooldown(rendId)
                end
			end
            return 100
		end,
		UpdateStatus = function()
            target_health = CalculateTargetHealthPct()
			s_ctime = CalculateCurrentTime()
		end,
		info = spells["Rend"].name,
	},
	cs = {
		id = function()
            return csId
        end,
		GetCD = function()
            if s1 ~= csId then
                local visible, _, _, _, _, _, expirationTime
                      = UnitAura("target", spells["Colossus Smash"].name, nil, "HARMFUL|PLAYER")
                local cooldown = GetCooldown(csId)
                local time_left = 0
                if visible ~= nil then
                    time_left = expirationTime - s_ctime - s_gcd
                end
                if time_left > cooldown then
                    cooldown = time_left
                end
				return cooldown
			end
			return 100
		end,
		UpdateStatus = function()
            target_health = CalculateTargetHealthPct()
			s_ctime = CalculateCurrentTime()
		end,
		info = spells["Colossus Smash"].name,
	},
	exedcreck = {
		id = function()
            return exeId
        end,
		GetCD = function()
			if IsUsableSpell(exeId) then
                local deadly_calm = UnitAura("player", spells["Deadly Calm"].name, nil, "HELPFUL|PLAYER")
                local recklessness = UnitAura("player", spells["Recklessness"].name, nil, "HELPFUL|PLAYER")
                if deadly_calm ~= nil or recklessness ~= nil then
                    return GetCooldown(exeId)
                end
			end
			return 100
		end,
		UpdateStatus = function()
            target_health = CalculateTargetHealthPct()
			s_ctime = CalculateCurrentTime()
		end,
		info = format([[%s: %s | %s]], spells["Execute"].name, spells["Deadly Calm"].name, spells["Recklessness"].name),
	},
    ms = {
		id = function()
            return msId
        end,
		GetCD = function()
            if s1 ~= msId then
                return GetCooldown(msId)
            end
            return 100
		end,
		UpdateStatus = function()
            target_health = CalculateTargetHealthPct()
			s_ctime = CalculateCurrentTime()
		end,
		info = spells["Mortal Strike"].name,
	},
	op = {
		id = function()
            return opId
        end,
		GetCD = function()
            if s1 ~= opId then
                if db.ignore_rend_op_less_than_20 == true and target_health < 20 then
                    return 100
                end
                local op_ready = UnitAura("player", spells["Taste for Blood"].name, nil, "HELPFUL|PLAYER")
                if op_ready ~= nil then
                    local op_cd = GetCooldown(opId)
                    if GetCooldown(msId) >= 1.5 then
                        return op_cd
                    end
                    return op_cd + 1.5 -- if MS is about to come off cd, prioritize it instead.
                end
            end
            return 100
        end,
		UpdateStatus = function()
            target_health = CalculateTargetHealthPct()
			s_ctime = CalculateCurrentTime()
		end,
		info = spells["Overpower"].name,
	},
	exe = {
		id = function()
            return exeId
        end,
		GetCD = function()
			if IsUsableSpell(exeId) then
                local exe_cd = GetCooldown(exeId)
                if GetCooldown(msId) >= 1.5 then
				    return exe_cd
                end
                return exe_cd + 1.5 -- if MS is about to come off cd, prioritize it instead.
			end
			return 100
		end,
		UpdateStatus = function()
            target_health = CalculateTargetHealthPct()
			s_ctime = CalculateCurrentTime()
		end,
		info = spells["Execute"].name,
	},
    slam = {
		id = function()
            return slamId
        end,
		GetCD = function()
            slam_cd = GetCooldown(slamId)
            local cs_timeleft = 100
            local visible, _, _, _, _, _, expirationTime
                      = UnitAura("target", spells["Colossus Smash"].name, nil, "HARMFUL|PLAYER")
            if visible ~= nil and GetCooldown(csId) <= 1.5 then
                cs_timeleft = expirationTime - s_ctime - s_gcd
            end
            if UnitPower("player") > 34 and GetCooldown(msId) >= 1.5 and cs_timeleft > 1.5 then
                return slam_cd
            end
            return slam_cd + 1.5 -- push slam back in priorities.
		end,
		UpdateStatus = function()
            target_health = CalculateTargetHealthPct()
			s_ctime = CalculateCurrentTime()
		end,
		info = spells["Slam"].name,
	},
}
mod.actions = actions
--------------------------------------------------------------------------------

function mod:UpdateQueue()
	q = {}
	for v in string.gmatch(db.prio, "[^ ]+") do
        if actions[v] then
            table.insert(q, v)
        else
            print("clcInfo", modName, "invalid action:", v)
        end
    end
	db.prio = table.concat(q, " ")
end

-- reads all the interesting data
local function GetStatus()
	-- current time
	s_ctime = GetTime()
	
	-- gcd value
	local start, duration = GetSpellCooldown(gcdId)
	s_gcd = start + duration - s_ctime
	s_gcdDur = duration
	if s_gcd < 0 then s_gcd = 0 end
end

local function GetNextAction()
	local n = #q
	
	-- parse once, get cooldowns, return first 0
	for i = 1, n do
		local action = actions[q[i]]
		local cd = action.GetCD()
		if debug.enabled then
			debug:AddBoth(q[i], cd)
		end
		if cd == 0 then
			return action.id(), q[i]
		end
		action.cd = cd
	end
	
	-- parse again, return min cooldown
	local minQ = 1
	local minCd = actions[q[1]].cd
	for i = 2, n do
		local action = actions[q[i]]
		if minCd > action.cd then
			minCd = action.cd
			minQ = i
		end
	end
	return actions[q[minQ]].id(), q[minQ]
end

function ArmsRotation()
	-- get the current db
	s1 = nil
	GetStatus()
	local action
	s1, action = GetNextAction()
	actions[action].UpdateStatus()
	
	s2, action = GetNextAction()
	
	return s1, s2
end


-- plug in
--------------------------------------------------------------------------------
local nextSkill
function emod.IconArmsCurrent(...)
	ArmsRotation(...)
	if nextSkill then nextSkill:DoUpdate() end
	return emod.IconSpell(s1, db.rangePerSkill or devName)
end
local function NextExec()
	return emod.IconSpell(s2, db.rangePerSkill or devName)
end
local function ExecCleanupNext()
	nextSkill = nil
end
function emod.IconArmsNext(...)
	nextSkill = emod.___e
	nextSkill:SetScript("OnUpdate", nil)
	nextSkill.exec = NextExec
	nextSkill.ExecCleanup = ExecCleanupNext
end

-- giving queue as command line
local function CmdArms(args)
	db.prio = table.concat(args, " ")
	mod:UpdateQueue()
	clcInfo:UpdateOptions()
end
clcInfo.cmdList["armsprio"] = CmdArms